from math import log, exp
import cPickle as pickle


def train(pos, neg, nwc=100, cname='model.pkl'):
    """ 
    This method trains a classifier.  The results will be written to a pickle 
    file named cname.
    
    pos and neg are the positive and negative training sets: lists of sets of
    features.
    """

    # just some shortcuts
    NP = len(pos)
    NN = len(neg)
    NX = NN + NP
    
    # features
    A = set()
    for dataset in (pos, neg):
        for datum in dataset:
            A.update(datum)
    
    features = []
    c0 = []
    c1 = []

    assert(NP > 0)
    assert(NN > 0)

    # set up the D vector
    D = [1.0/NX] * NX
    epsilon = 1.0/NX

    for w in range(nwc):
        bestZ = 10e10
        bestFeat = 0
        bestneg_W0 = 0
        bestneg_W1 = 0
        bestpos_W0 = 0
        bestpos_W0 = 0

        # get sums of D for +ve and -ve
        Dpos = sum(D[0:NP])
        Dneg = sum(D[NP:NX])

        # check each feature and use the best one
        for a in A:
            pos_W0 = 0.0
            neg_W0 = 0.0
            pos_W1 = 0.0
            neg_W1 = 0.0

            # compute W1s
            for i in range(NP): 
                if a in pos[i]: 
                    pos_W1 += D[i] 
            for i in range(NN): 
                if a in neg[i]: 
                    neg_W1 += D[i+NP] 

            # W0s are just leftover D-weights
            pos_W0 = Dpos - pos_W1
            neg_W0 = Dneg - neg_W1

            # now we can get our "badness" measure Z
            Z =  pos_W0 * ((neg_W0 + epsilon) / (pos_W0 + epsilon))**0.5 
            Z += neg_W0 * ((pos_W0 + epsilon) / (neg_W0 + epsilon))**0.5
            Z += pos_W1 * ((neg_W1 + epsilon) / (pos_W1 + epsilon))**0.5 
            Z += neg_W1 * ((pos_W1 + epsilon) / (neg_W1 + epsilon))**0.5

            if Z < bestZ : 
                bestFeat = a
                bestZ = Z
                bestneg_W0 = neg_W0
                bestneg_W1 = neg_W1
                bestpos_W0 = pos_W0
                bestpos_W1 = pos_W1

        # alright, we have a weak classifier
        features.append(bestFeat)
        c0.append(log((bestpos_W0 + epsilon) / (bestneg_W0 + epsilon)))
        c1.append(0.5 * log((bestpos_W1 + epsilon) / (bestneg_W1 + epsilon)))

        # now update
        h = [c0[-1]] * NX
        for i in range(NP): 
            if bestFeat in pos[i]: h[i] = c1[-1]
        for i in range(NN): 
            if bestFeat in neg[i]: h[NP+i] = c1[-1] 

        # adapt D based on h
        D = [D[i] * exp(-1.0 * ((i < NP) and 1.0 or -1.0) * h[i]) for i in range(NX)]
        D = [elem/sum(D) for elem in D]
        
        print len(c0), ' = if', features[-1],'then', c1[-1], 'else', c0[-1]
        
    
    pickle.dump((c0, c1, features), open(cname, 'w'))


def test(X, model='model.pkl'):
    """
    Return the hypothesis vector: the +1/-1 class for each datum in the
    validation set X.
    """
    c0, c1, features = pickle.load(open(model))
    
    hypothesis = []
    
    for x in X:
        h = 0
        for i in xrange(len(features)):
            if features[i] in x:
                h += c1[i]
            else:
                h += c0[i]
        hypothesis.append(1 if h > 0 else -1)
        
    return hypothesis
    
    
def loaddata(fname):
    data = []
    for line in open(fname).readlines():
        data.append(set(int(f) for f in line.strip().split(',')))
    return data


if __name__ == "__main__":
   
    training = False
    validating = True
    
    if training:
        # load the training data
        postrain = loaddata('pos_train.txt')
        negtrain = loaddata('neg_train.txt')
    
        # train
        train(postrain, negtrain)

    if validating:
        # false negative rate
        posvalidate = loaddata('pos_validate.txt')
        fnr = sum(1.0 for h in test(posvalidate) if h < 0.) / len(posvalidate)
    
        # false positive rate
        negvalidate = loaddata('neg_validate.txt')
        fpr = sum(1.0 for h in test(negvalidate) if h >= 0.) / len(negvalidate)
    
        print '============================'
        print 'false negative rate = %.4f (over %d positive data)' % (fnr, len(posvalidate))
        print 'false positive rate = %.4f (over %d negative data)' % (fpr, len(negvalidate))
    
    